"""Entry point for nonadditivity package."""

import click

from nonadditivity import __version__ as version
from nonadditivity.nonadditivity_workflow import run_nonadd_calculation
from nonadditivity.utils.commandline import InputOptions, add_entry_point_options
from nonadditivity.utils.log import get_logger, setup_logging

logger = get_logger()


@click.command("Nonadditivity CLI", context_settings={"show_default": True})
@click.version_option(version=version)
@add_entry_point_options
def main(input_options: InputOptions) -> None:
    """Entry point for the nonnaditivity command line tool.

    Nonadditivity Analysis based on matched MMPs

    Args:
      input_options (InputOptions): parameters for the program.
    """
    setup_logging(level=input_options.verbose, out_file=input_options.log_file)
    run_nonadd_calculation(input_options=input_options)


if __name__ == "__main__":
    main()  # pylint: disable=E1120



"""Nonadditivity Analysis.

Copyright (c) 2024, F. Hoffmann-La Roche Ltd.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

   * Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above
     copyright notice, this list of conditions and the following
     disclaimer in the documentation and/or other materials provided
     with the distribution.
   * Neither the name of F. Hoffmann-La Roche Ltd. nor the names of
     its contributors may be used to endorse or promote products
     derived from this software without specific prior written
     permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import logging
from importlib.metadata import version
from logging import NullHandler

from rdkit import rdBase

from nonadditivity.classification import classify
from nonadditivity.utils.commandline import add_entry_point_options
from nonadditivity.workflow import (
    parse_input_file,
    run_mmpdlib_code,
    run_nonadditivity_core,
    write_output_files,
    write_smiles_id_file,
)

__version__ = version("nonadditivity")

rdBase.DisableLog("rdApp.*")  # pylint: disable=I1101

logging.getLogger(__name__).addHandler(logging.NullHandler())
del NullHandler



"""Custom errors to be thrown when running mmpdb."""


class FragmentationError(Exception):
    """Thrown when mmpdb fragment fails."""


class IndexingError(Exception):
    """Thrown when mmpdb index fails."""



"""Type aliases for rdkit Molecules and Atoms."""

from typing import TypeAlias

from rdkit.Chem import rdchem

Molecule: TypeAlias = rdchem.Mol  # pylint:disable=I1101
Atom: TypeAlias = rdchem.Atom  # pylint:disable=I1101



"""Implement compound classification.

Implements all classification functions on a compound level
that are called in the Compound class.
All these functions are higly specific, do not use them outside
of their intended application.
"""

from rdkit import Chem, DataStructs
from rdkit.Chem import Crippen, GraphDescriptors, rdMolDescriptors

from nonadditivity.utils.types import Molecule


def get_aromatic_indices(**kwargs: Molecule) -> set[int]:
    """Get list of indices for all aromatic atoms in molecule.

    use keyword argument 'molecule' for this function to work.

    Returns:
        set[int]: indices of aromatic atoms in molecule
    """
    mol = kwargs["molecule"]
    return {
        i for i in range(mol.GetNumAtoms()) if mol.GetAtomWithIdx(i).GetIsAromatic()
    }


def get_fused_ring_indices(**kwargs: Molecule) -> list[set[int]]:
    """Detects the number of fused rings in the given molecule.

    Returns:
        list[set[int]]: list of sets with fused ring system indices.
    """
    aromatic_atoms = get_aromatic_indices(**kwargs)
    sssr = Chem.GetSSSR(kwargs["molecule"])
    fused_rings = []
    for outer_i, outer in enumerate(sssr):
        if not all(i in aromatic_atoms for i in outer):
            continue
        newring = set(outer)
        same_length = False
        while not same_length:
            old_length = len(newring)
            for _, inner in enumerate(sssr[outer_i + 1 :]):
                if not all(i in aromatic_atoms for i in inner):
                    continue
                if len(newring.intersection(inner)) >= 2:
                    newring.update(inner)
            same_length = old_length == len(newring)
        if newring == set(outer):
            continue
        if not any(len(i.intersection(newring)) == len(newring) for i in fused_rings):
            fused_rings.append(newring)
    return fused_rings


def get_num_heavy_atoms(**kwargs: Molecule) -> int:
    """Get number of heavy atoms in a molecule.

    Returns number of heavy atoms of the compound
    using rdkits Mol.GetNumHeavyAtoms()

    use keyword argument 'molecule' for this function to work.


    Returns:
        int: Number of heavy atoms.
    """
    molecule = kwargs["molecule"]
    return molecule.GetNumHeavyAtoms()


def get_num_stereo_centers(**kwargs: Molecule) -> int:
    """Get number of stereocenter in a molecule.

    Returns number of stereo Centers of the compound
    using rdkits new implementation of
    Chem.FindMolChiralCenters()

    use keyword argument 'molecule' for this function to work.


    Returns:
        int: Number of stereo centers.
    """
    molecule: Molecule = kwargs["molecule"]
    return len(
        Chem.FindMolChiralCenters(
            mol=molecule,
            includeUnassigned=True,
            useLegacyImplementation=False,
        ),
    )


def has_unassigned_stereocenters(**kwargs: Molecule) -> bool:
    """Check if a molecule has unassigned stereocenters.

    use keyword argument 'molecule' for this function to work.

    Returns:
        bool: True if mol has unassigned stereocenter
    """
    molecule: Molecule = kwargs["molecule"]
    centers: list[str] = [
        v[1]
        for v in Chem.FindMolChiralCenters(
            mol=molecule,
            includeUnassigned=True,
        )
    ]
    return any(c.lower() == "?" for c in centers)


def get_num_h_bond_acceptors(**kwargs: Molecule) -> float:
    """Calculate number of hbond acceptors for the molecule.

    use keyword argument 'molecule' for this function to work.

    Returns:
        float: numn h bond acceptors
    """
    return rdMolDescriptors.CalcNumHBA(  # pylint: disable=I1101
        kwargs["molecule"],
    )


def get_num_h_bond_donors(**kwargs: Molecule) -> float:
    """Calculate number of hbond donors for the molecule.

    use keyword argument 'molecule' for this function to work.

    Returns:
        float: numn h bond donors
    """
    return rdMolDescriptors.CalcNumHBD(  # pylint: disable=I1101
        kwargs["molecule"],
    )


def get_charge(**kwargs: Molecule) -> int:
    """Calculate charge of a molecule.

    use keyword argument 'molecule' for this function to work.

    Returns:
        float: charge
    """
    return Chem.GetFormalCharge(  # type: ignore pylint: disable=E1101
        kwargs["molecule"],
    )


def get_polar_surface_area(**kwargs: Molecule) -> float:
    """Calculate the topological polar surface areas for a molecule.

    use keyword argument 'molecule' for this function to work.

    Returns:
        float: polar surface area
    """
    return rdMolDescriptors.CalcTPSA(kwargs["molecule"])  # pylint: disable=I1101


def get_log_p(**kwargs: Molecule) -> list[float]:
    """Calculate log_p for the molecule.

    use keyword argument 'molecule' for this function to work.

    Returns:
        list[float]: log_p values
    """
    return Crippen.MolLogP(kwargs["molecule"])


def get_num_sp3_carbons(**kwargs: Molecule) -> int:
    """Calculate number of sp3 carbons in molecule.

    use keyword argument 'molecule' for this function to work.

    Returns:
        list[float]: fraction of sp3 carbons
    """
    return sum(
        a.GetHybridization() == Chem.HybridizationType.SP3 and a.GetSymbol() == "C"  # pylint: disbale=E1101
        for a in kwargs["molecule"].GetAtoms()
    )


def get_num_rotatable_bonds(**kwargs: Molecule) -> int:
    """Calculate number of rotatable bonds for a molecule transformation.

    use keyword argument 'molecule' for this function to work.

    Returns:
        list[float]: num rotatable bonds
    """
    # molecule = Chem.AddHs(kwargs["molecule"])
    return rdMolDescriptors.CalcNumRotatableBonds(
        kwargs["molecule"],
    )  # pylint: disable=I1101


def get_chi0(**kwargs: Molecule) -> list[float]:
    """Calculate chi0 for both parts of the transformation.

    use keyword argument 'molecule' for this function to work.

    Returns:
        list[float]: Chi0 values
    """
    return GraphDescriptors.Chi0(kwargs["molecule"])


def get_chi1(**kwargs: Molecule) -> list[float]:
    """Calculate chi1 for both parts of the transformation.

    use keyword argument 'molecule' for this function to work.

    Returns:
        list[float]: chi1 values
    """
    return GraphDescriptors.Chi1(kwargs["molecule"])


def get_chi2(**kwargs: Molecule) -> list[float]:
    """Calculates chi2 for both parts of the transformation.

    use keyword argument 'molecule' for this function to work.

    Returns:
        list[float]: chi2 values
    """
    return GraphDescriptors.Chi2n(kwargs["molecule"])


def get_morgan_fp(
    **kwargs: Molecule,
) -> DataStructs.cDataStructs.UIntSparseIntVect:  # pylint: disable=I1101
    """Calculate the morgan fingerprint with radius 2 for a molecule.

    use keyword argument 'molecule' for this function to work.


    Returns:
        DataStructs.cDataStructs.UIntSparseIntVect: MorganFP with radius 2
    """
    return rdMolDescriptors.GetMorganFingerprint(  # type: ignore pylint:disable=E1101
        mol=kwargs["molecule"],
        radius=2,
    )



"""Implement classification on all three levels.

Classification submodule of the NonadditivityAnalysis module. All Functions and objects
used here are higly specific. Don't use them on their own.
"""

import pandas as pd
from tqdm.auto import tqdm

from nonadditivity.classification import Circle, Compound, Transformation


def _create_compound_dict(
    per_compound_dataframe: pd.DataFrame,
) -> dict[str, Compound]:
    """Create dict mapping id to compound object.

    Creates dict mapping str to Compound objects for every entry in
    per_compound_dataframe.

    Args:
        per_compound_dataframe (pd.DataFrame): Dataframe w/ 'RDKit_Moleculs' & 'SMILES'

    Returns:
        dict[str, Compound]: dict mapping str to compound objects
    """
    return {
        i: Compound(molecule=m, compound_id=i, smiles=s)
        for m, i, s in zip(
            per_compound_dataframe.RDKit_Molecules.to_numpy(),
            per_compound_dataframe.index.to_numpy(),
            per_compound_dataframe.SMILES.to_numpy(),
        )
    }


def _update_per_compound_dataframe(
    per_compound_dataframe: pd.DataFrame,
    compound_dict: dict[str, Compound],
) -> None:
    """Add classification details to per compound dataframe.

    Adding 'num_stereocenters' and 'num_ortho_configurations' and
    'has_unassigned_stereocenters' for every compound to the per compound dataframe.

    Args:
        per_compound_dataframe (pd.DataFrame): dataframe to add info to
        compound_dict (dict[str, Compound]): dictionary containing all compounds
        present in per_compound_dataframe.
    """
    props = Compound.Properties

    for compound_property, compound_property_str in (
        pbar := tqdm(
            Compound.classification_keys.items(),
            total=len(Compound.classification_keys.items()),
            desc="Calculating Compound properties",
            ascii="░▒█",
        )
    ):
        if compound_property in (
            props.NUM_HEAVY_ATOMS,
            props.MORGANFP,
            props.ORTHO_INDICES,
            props.NUM_ORTHO_CONFIGURATIONS,
        ):
            continue
        pbar.set_description(
            desc=(
                "Calculating Compound properties: "
                f"{compound_property_str.replace('_',' ')}"
            ),
        )
        values = []
        for key in tqdm(per_compound_dataframe.index.to_numpy()):
            values.append(compound_dict[key].get_property(compound_property))
        per_compound_dataframe[compound_property_str] = values


def classify_compounds(
    per_compound_dataframe: pd.DataFrame,
) -> dict[str, Compound]:
    """Classify all compounds in per compound dataframe and create compound instances.

    Creates Compound objects for every entry in the per_compound_dataframe and adds
    columns "num_stereocenters" and "num_ortho_patterns to the per_compound_dataframe.

    Args:
        per_compound_dataframe (pd.DataFrame): df containing compounds

    Returns:
        dict[str, Compound]: dict mapping compound id to compound object
    """
    compound_dict = _create_compound_dict(
        per_compound_dataframe=per_compound_dataframe,
    )
    _update_per_compound_dataframe(
        per_compound_dataframe=per_compound_dataframe,
        compound_dict=compound_dict,
    )
    return compound_dict


def _get_na_compounds(
    na_dataframe: pd.DataFrame,
    compound_dict: dict[str, Compound],
) -> list[list[Compound]]:
    """Get a list of compounds for every row in the naa dataframe.

    Returns list lists of compounds that are present in the nonadditivity circles
    in the na_dataframe.

    Args:
        na_dataframe (pd.DataFrame): dataframe containint nonadditivity circles
        compound_dict (dict[str, Compound]): dict mapping compound id to compound object

    Returns:
        list[list[Compound]]: list of lists with compounds in na_dataframe
    """
    return [
        [compound_dict[i] for i in ids]
        for ids in zip(
            na_dataframe.Compound1.to_numpy(),
            na_dataframe.Compound2.to_numpy(),
            na_dataframe.Compound3.to_numpy(),
            na_dataframe.Compound4.to_numpy(),
        )
    ]


def _get_transformations_for_na_dataframe(
    na_compounds: list[list[Compound]],
    per_compound_dataframe: pd.DataFrame,
) -> list[tuple[Transformation, ...]]:
    """Get transformation objects for transformations in na dataframe.

    Returns a list of tuples of transformation, so a Transformation object for
    every transformation in a circle of the na_dataframe is generated:

    CPD1 ======= T1 ====== > CPD2
      |                        |
      |                        |
     T2                       T3
      |                        |
      /                        /
    CPD4 ======= T4 ====== > CPD3

    Args:
        na_compounds (list[list[Compound]]): list of lists for compounds in\
            naa dataframe circles
        per_compound_dataframe (pd.DataFrame): per compound dataframe

    Returns:
        list[tuple[Transformation, ...]]:\
            list of tuples containing transformations for all the naa circles.
    """
    return [
        tuple(
            Transformation(
                compound_1=c[i],
                compound_2=c[j],
                constant_smarts=per_compound_dataframe.loc[
                    c[i].compound_id,
                    "Neighbor_dict",
                ][c[j].compound_id][1],
                transformation_smarts=per_compound_dataframe.loc[
                    c[i].compound_id,
                    "Neighbor_dict",
                ][c[j].compound_id][0],
            )
            for i, j in ((0, 1), (0, 3), (1, 2), (3, 2))
        )
        for c in tqdm(
            na_compounds,
            total=len(na_compounds),
            desc="Classifying Transformation",
            ascii="░▒█",
        )
    ]


def _create_circle_objects(
    na_transformations: list[tuple[Transformation, ...]],
) -> list[Circle]:
    """Return list of Circle objects.

    Args:
        na_transformations (list[ tuple[Transformation, ...] ]): \
            list of tuples where every\
                tuple has the four transformations makeing up a circle.

    Returns:
        list[Circle]: Circle objects.
    """
    return [
        Circle(
            transformation_1=transformation_1,
            transformation_2=transformation_2,
            transformation_3=transformation_3,
            transformation_4=transformation_4,
        )
        for (
            transformation_1,
            transformation_2,
            transformation_3,
            transformation_4,
        ) in na_transformations
    ]


def _classify_ortho(circle: Circle) -> str:
    """Create ortho classification string.

    Melts the two classifications HAS_ORTHO_SUBSTITUENT_INTRODUCED
    and HAS_TRANSFORMATION_AT_ORTHO into one so ortho_classification
    can either be 'Both', 'Introduced', 'Changed' or 'None'.

    Args:
        circle (Circle): Circle to classify

    Returns:
        str: 'None', 'Both', 'Introduced' or 'Changed'.
    """
    if circle.get_property(
        circle.Properties.HAS_ORTHO_SUBSTITUENT_INTRODUCED,
    ) and circle.get_property(circle.Properties.HAS_TRANSFORMATION_AT_ORTHO):
        return "Both"
    if circle.get_property(circle.Properties.HAS_ORTHO_SUBSTITUENT_INTRODUCED):
        return "Introduced"
    if circle.get_property(circle.Properties.HAS_TRANSFORMATION_AT_ORTHO):
        return "Changed"
    return "None"


def _classify_circles(
    circles: list[Circle],
    property_key: Circle.Properties,
) -> list[str] | list[float] | list[tuple[int]] | list[bool]:
    """Gets the classification for <property_key> for a list of circles.

    Args:
        circles (list[Circle]): circles to classify.
        property_key (Circle.Properties): property to get.

    Returns:
        list[str] | list[float] | list[tuple[int]] | list[bool]:
        classification for circles and property_key
    """
    classification: list[str] | list[float] | list[tuple[int]] = []  # type:ignore
    for circle in circles:
        classification.append(circle.get_property(property_key))  # type: ignore
    return classification


def _is_surprisingly_nonadditive(circle: Circle) -> bool:
    """Check whether a circle's nonadditivity is surprising.

    Use keyword "circle" for this to work.

    A circle is considered surprising i fnone of the following is true:
    Distance between R groups ≤ 2 atoms
    Tanimoto similarity of the transformation < 0.4
    Number of exchanged heavy atoms > 10
    Linker exchange transformations
    Transformations with unassigned or inverted stereocenters

    Args:
        circle (Circle): circle to check.

    Returns:
        bool: True if surprising false if Mundane.
    """
    if circle.get_property(Circle.Properties.DISTANCE_BETWEEN_R_GROUPS) <= 2:
        return False
    if circle.get_property(Circle.Properties.MIN_TANIMOTO) < 0.4:
        return False
    if circle.get_property(Circle.Properties.NUM_HEAVY_ATOMS_DIFF) > 10:
        return False
    if circle.get_property(Circle.Properties.MAX_NUM_MMPDB_CUTS) > 1:
        return False
    return (
        circle.get_property(Circle.Properties.HAS_INVERSION_IN_TRANSFORMATION) == "None"
    )


def _update_na_dataframe(
    na_dataframe: pd.DataFrame,
    circles: list[Circle],
) -> None:
    """Add circle classification columns to the na_dataframe.

    Args:
        na_dataframe (pd.DataFrame): dataframe to add columns to.
        circles (list[Circle]): list of circle objects in the order of them occuring\
            in the na_dataframe.
    """
    props = Circle.Properties
    for circle_property, circle_property_str in (
        pbar := tqdm(
            Circle.classification_keys.items(),
            total=len(Circle.classification_keys.items()),
            desc="Calculating Circle properties",
            ascii="░▒█",
        )
    ):
        if circle_property in (
            props.DISTANCE_BETWEEN_R_GROUPS,
            props.HAS_ORTHO_SUBSTITUENT_INTRODUCED,
            props.HAS_TRANSFORMATION_AT_ORTHO,
        ):
            continue
        pbar.set_description(
            desc=(
                "Calculating Circle properties: "
                f"{circle_property_str.replace('_',' ')}"
            ),
        )
        na_dataframe[circle_property_str] = _classify_circles(
            circles=circles,
            property_key=circle_property,
        )

    na_dataframe["ortho_classification"] = [_classify_ortho(c) for c in circles]
    na_dataframe["num_atoms_between_r_groups"] = _classify_circles(
        circles=circles,
        property_key=props.DISTANCE_BETWEEN_R_GROUPS,
    )
    na_dataframe["classification"] = [
        "surprising" if _is_surprisingly_nonadditive(c) else "mundane" for c in circles
    ]


def classify_circles(
    na_transformations: list[tuple[Transformation, ...]],
    na_dataframe: pd.DataFrame,
) -> list[Circle]:
    """Create circle objects and classify them.

    Creates Circle objects for every entry in the na_dataframe and classifies
    them. Then adds the classification to the na_dataframe.

    Args:
        na_transformations (list[ tuple[Transformation, ...] ]):
            na_transformations the circles in na_dataframe are made of.
        na_dataframe (pd.DataFrame): na_dataframe with circles to classify.

    Returns:
        list[Circle]: list of classified circle objects.
    """
    circles = _create_circle_objects(na_transformations=na_transformations)
    _update_na_dataframe(na_dataframe=na_dataframe, circles=circles)
    return circles


def classify(
    per_compound_dataframe: pd.DataFrame,
    na_dataframe: pd.DataFrame,
) -> tuple[list[list[Compound]], list[tuple[Transformation, ...]], list[Circle]]:
    """Classify compounds, circles and transformations in the circle.

    Classifies all compound in the per_compound_dataframe and then classifies all
    the circles in the na_dataframe.

    Args:
        per_compound_dataframe (pd.DataFrame): dataframe containing compounds found\
            in the nonadditivity circles.
        na_dataframe (pd.DataFrame): dataframe containing circles to be classified.

    Returns:
        tuple[ list[list[Compound]], list[tuple[Transformation, ...]], list[Circle]]:\
             tuple containing\
            list of lists of compounds found in the nonadditivity circles,\
            list of tuples of transformations found in the nonadditivty circles\
            and list of circles in the na_dataframe.
    """
    compound_dict = classify_compounds(
        per_compound_dataframe=per_compound_dataframe,
    )
    na_compounds = _get_na_compounds(
        na_dataframe=na_dataframe,
        compound_dict=compound_dict,
    )
    na_transformations = _get_transformations_for_na_dataframe(
        na_compounds=na_compounds,
        per_compound_dataframe=per_compound_dataframe,
    )
    na_circles = classify_circles(
        na_transformations=na_transformations,
        na_dataframe=na_dataframe,
    )
    return na_compounds, na_transformations, na_circles



"""Implements dtc classification in nonadditivity package."""

from nonadditivity.classification.classification_classes.circle import Circle
from nonadditivity.classification.classification_classes.compound import Compound
from nonadditivity.classification.classification_classes.transfromation import (
    Transformation,
)
from nonadditivity.classification.classify import classify

__all__ = ["classify", "Circle", "Compound", "Transformation"]



"""Parsing input in nonadditivity analysis package.

The Functions in this file handle all the file reading and input parsing for the
NonadditivityAnalysis package.
"""

import os.path
from collections.abc import Sequence
from pathlib import Path

import numpy as np
import pandas as pd
from rdkit import Chem, RDConfig
from rdkit.Chem import Descriptors, SaltRemover
from tqdm.auto import tqdm

from nonadditivity.utils.log import get_logger
from nonadditivity.utils.math import is_number

logger = get_logger()
DELIMITERS: dict[str, str | None] = {
    "comma": ",",
    "tab": "\t",
    "space": " ",
    "semicolon": ";",
    "": None,
}
LOG_FLAGS = [
    "pIC50",
    "pEC50",
    "pKi",
    "pKd",
    "pCC50",
    "pIC20",
    "pID50",
    "PCHEMBL",
]
DEFAULT_UNIT = ["M"]
CONVERSIONS = {
    "m": (1.0, 1.0),
    "mm": (1e-3, 3.0),
    "um": (1e-6, 6.0),
    "nm": (1e-9, 9.0),
    "pm": (1e-12, 12.0),
}


def _read_in_data(
    infile: Path,
    property_columns: list[str],
    delimiter: str = "tab",
    series_column: str | None = None,
) -> pd.DataFrame:
    """Read data from infile and converts it to a pandas Dataframe.

    It guesses the SMILES and ID column with substring search.
    It renames the columns as follows:
        > Any column containing 'SMILES' is renamed to -> 'SMILES'
        > Any column containng 'ID' or 'SRN' and that is not specified by series_column
            is renamed to -> 'ID'.
        > the column given in series_column is renamed to -> 'SERIES'

    Args:
        infile (str): file to read structures from.
        property_columns (list[str]): Names of the property columns.
        delimiter (str, optional): delimiter used in input file. Defaults to "tab".
        series_column (str | None, optional): name of the series column. Defaults to
        None.

    Raises:
        ValueError: If more than one smiles column is found in the input file.
        ValueError: If more than one column is found with id or srn in it that is not
        the series_column.

    Returns:
        pd.DataFrame: parsed dataframe.
    """
    input_dataframe = pd.read_table(
        filepath_or_buffer=infile,
        sep=DELIMITERS[delimiter],
    )
    smiles, ids = [], []

    for column in input_dataframe.columns.to_numpy():
        if "smiles" in column.lower():
            logger.info("Smiles Column found: %s", column)
            smiles.append(column)
        if "srn" in column.lower() or "id" in column.lower():
            logger.info("Identifier Column found: %s", column)
            ids.append(column)

    for property_column in property_columns:
        if property_column in input_dataframe.columns:
            logger.info("Property Column found: %s", property_column)
            continue
        raise ValueError(
            f"Given Property Column not found in input file: {property_column}",
        )

    if series_column:
        input_dataframe = input_dataframe.rename(columns={series_column: "Series"})
        logger.info("Series Column found: %s", series_column)

    if len(smiles) == 1:
        input_dataframe = input_dataframe.rename(columns={smiles[0]: "SMILES"})
    else:
        raise ValueError(
            "Please assign unambiguous names (no overlap in 'SMILES', 'ID', 'SRN'.",
        )
    input_dataframe = input_dataframe.rename(columns={ids[0]: "Compound_ID"})
    input_dataframe = input_dataframe.set_index("Compound_ID", drop=False)
    return input_dataframe


def _get_salt_remover(
    salt_file: str = os.path.join(RDConfig.RDDataDir, "Salts.txt"),
) -> SaltRemover.SaltRemover:
    """Instaciates a rdkit SaltRemover object.

    Give a path to a file with salt_definitions, otherwise the default
    Salts.txt file from rdkit is used.

    Args:
        salt_file (str, optional): Path to file with salt definitions. Defaults to
        os.path.join( RDConfig.RDDataDir, "Salts.txt", ).

    Returns:
        SaltRemover.SaltRemover: rdkit Salt Remover object.
    """
    return SaltRemover.SaltRemover(defnFilename=salt_file)


def _parse_input_smiles(
    dataframe: pd.DataFrame,
    verbose: bool = True,
) -> tuple[list[str], list[str], pd.DataFrame]:
    """Parse smiles in input file.

    Takes a dataframe and adds columns containing molecular weight and number of heavy
    atoms for the molecule specified in the smiles_column. Prior to the property
    calculation all salts are removed. The SMILES in the  smiles_column are replaced
    with SMILES not containing salt. If a SMILES could not be parsed, mol_weighs and
    num_heavies is set to -1 for that entry. Returns a list with SMILES that have
    unknown salts, and a list of smiles that could not be parsed by rdkits
    MolFromSmiles function.

    Args:
        dataframe (pd.DataFrame): Dataframe to add the properties.
        verbose (bool): wheter to display tqdm status bar. Defaults to True.

    Returns:
        tuple[list[str], ...]: lists with SMILES containing unknown salts or unparsable
        SMILES.
    """
    (new_smiles, mol_weights, num_heavies, unknown_salts, faulty_smiles, mols) = (
        [],
        [],
        [],
        [],
        [],
        [],
    )
    remover = _get_salt_remover()
    for index, original_smiles in tqdm(
        zip(dataframe.index.to_numpy(), dataframe.SMILES.to_numpy()),
        total=len(dataframe),
        desc="Parsing Input SMILES",
        ascii="░▒█",
        disable=not verbose,
    ):
        try:
            smiles = original_smiles.replace("\\\\", "\\")
            mol = Chem.MolFromSmiles(  # type: ignore pylint: disable=E1101
                smiles,
                sanitize=True,
            )
            mol = remover.StripMol(mol)
            smiles = Chem.MolToSmiles(  # type: ignore pylint: disable=E1101
                mol,
                canonical=True,
            )
            if "." in smiles:
                unknown_salts.append(index)
                mols.append(-1)
                num_heavies.append(-1)
                mol_weights.append(-1)
                new_smiles.append(smiles)
                continue
            num_heavies.append(mol.GetNumHeavyAtoms())
            mol_weights.append(Descriptors.MolWt(mol))
            mols.append(mol)
            new_smiles.append(smiles)
        except Exception:  # pylint: disable=W0703
            faulty_smiles.append(index)
            num_heavies.append(-1)
            mol_weights.append(-1)
            mols.append(-1)
            new_smiles.append(original_smiles)
            continue
    dataframe["SMILES"] = new_smiles
    dataframe["Molecular_Weight"] = mol_weights
    dataframe["Num_Heavy_Atoms"] = num_heavies
    dataframe["RDKit_Molecules"] = mols
    logger.info("SMILES parsed.")
    return unknown_salts, faulty_smiles, dataframe


def _remove_unknown_and_faulty_structures(
    dataframe: pd.DataFrame,
    unknown_smiles_indices: list[str],
    faulty_smiles_indices: list[str],
    error_path: Path | None = None,
) -> pd.DataFrame:
    """Remove structures with unparsable smiles.

    Removes entries from the Dataframe with the SMILES given in unknown_smiles_indices
    and faulty_smiles_indices. All takes plac in place.

    Args:
        dataframe (pd.DataFrame): dataframe to check
        unknown_smiles_indices (list[int]): smiles of unknown structures
        faulty_smiles_indices (list[int]): smiles of faulty structures
        error_path (str): path to write erronous smiles to.
    """
    if unknown_smiles_indices:
        logger.warning(
            "%i Molecules with unknown salts were found and "
            "will be excluded from further analysis",
            len(unknown_smiles_indices),
        )
    if faulty_smiles_indices:
        logger.warning(
            "%i Molecules with invalid smiles were found and "
            "will be excluded from further analysis",
            len(faulty_smiles_indices),
        )
    if any(unknown_smiles_indices) or any(faulty_smiles_indices):
        logger.warning("Check %s for erronous entries.", str(error_path))
    if error_path:
        faulty_dataframe = dataframe[
            dataframe.index.isin(
                faulty_smiles_indices + unknown_smiles_indices,
            )
        ]
        faulty_dataframe.to_csv(
            path_or_buf=error_path,
            sep="\t",
            columns=["SMILES"],
            index=False,
            header=False,
        )
    return dataframe.drop(
        index=np.unique(
            unknown_smiles_indices + faulty_smiles_indices,  # type: ignore
        ),
    )


def _check_for_censored_values(
    dataframe: pd.DataFrame,
    property_columns: list[str],
) -> pd.DataFrame:
    """Check dataframe for censored values.

    Checks for values that begin with either *, < or >. if such a value is found, the
    sign is removed from the value and stored in a seperat CENSORED column in the
    dataframe.

    Args:
        dataframe (pd.DataFrame): Dataframe to check
        property_columns (str): property row to check
    """
    dataframe = dataframe.fillna("notanumber")
    for property_column in property_columns:
        censored = []
        for index, value in zip(
            dataframe.index.to_numpy(),
            dataframe[property_column].to_numpy(),
        ):
            if is_number(value):
                censored.append("")
                continue
            if value[0] in (">", "<", "*") and is_number(value[1:]):
                dataframe.loc[index, property_column] = value[1:]
                censored.append(value[0])
                continue
            censored.append("NA")
            dataframe.loc[index, property_column] = "0"
        dataframe[f"{property_column}_Censors"] = censored
    return dataframe


def _remove_duplicate_structures(dataframe: pd.DataFrame) -> pd.DataFrame:
    """Remove structures that have equal smiles, keeping the first entry.

    Checks the dataframe for entries that have the same value in their
    SMILES COLUMN. It drops all but the first in place. It also checks
    for duplicates in the ID column, if found, an Value Error is raised.

    Args:
        dataframe (pd.DataFrame): DataFrame to check.

    Raises:
        ValueError: Thrown if two or more entries have the same ID.
    """
    dataframe = dataframe.drop_duplicates(subset="SMILES", keep="first")
    if not dataframe.index.is_unique:
        raise ValueError(
            f"Two or more entries for the same identifier: \
                {dataframe.index[np.where(dataframe.index.duplicated())[0].item()]}",
        )
    logger.info(
        (
            "Entries with duplicate SMILES were removed, "
            "the first occuring entry was kept."
        ),
    )
    return dataframe


def _check_and_convert_activities(
    dataframe: pd.DataFrame,
    property_columns: list[str],
    units: list[str | None],
) -> tuple[list[str], pd.DataFrame]:
    """Assert values possible and convert them according to units given.

    Checks dataframes property_columns column and converts, if needed, the units to
    molar and takes the negative logarithm of it. The values are stored in a new
    column. property_columns and units either have to be both str, or both lists with
    the same length.

    Args:
        dataframe (pd.DataFrame): Dataframe with the values to be converted
        property_columns (Union[str, list[str]]): Name(s) of the property columns to be
        converted.
        units (Union[str, list[str]]): Unit(s) of the values in the property column(s)

    Returns:
        list[str]: Columnames where the converted values are stored.
    """

    def convert(
        dataframe: pd.DataFrame,
        property_column: str,
        unit: str | None,
    ) -> tuple[pd.DataFrame, str]:
        """Convert values.

        Converts values in property column to molar -log10.
        unit is the original unit.

        Args:
            dataframe: (pd.DataFrame):
            property_column (str): name of column with activities
            unit (str): unit of the activities.

        Returns:
            str: name of the new column with the converted values.
        """
        try:
            dict_unit = unit.lower()  # type: ignore
        except AttributeError:
            dict_unit = "m"
        if not any(
            log_flag.lower() in property_column.lower() for log_flag in LOG_FLAGS
        ):
            dataframe = dataframe[dataframe[property_column] > 0]
            dataframe = dataframe.assign(
                **{
                    f"p{property_column}": -np.log10(
                        dataframe[property_column] * CONVERSIONS[dict_unit][0],
                    ),
                },
            )
            return dataframe, f"p{property_column}"
        if CONVERSIONS[dict_unit] != 1:
            dataframe[property_column] = (
                dataframe[property_column] + CONVERSIONS[dict_unit][1]
            )
        return dataframe, property_column

    columnames = []
    for index, property_column in enumerate(property_columns):
        dataframe[property_column] = dataframe[property_column].astype(float)
        if "pchembl" in property_column.lower():
            columnames.append(property_column)
            continue
        if units[index] == "noconv":
            columnames.append(property_column)
        else:
            dataframe, newcol = convert(
                dataframe=dataframe,
                property_column=property_column,
                unit=units[index],
            )
            columnames.append(newcol)
        dataframe = dataframe.rename(
            columns={
                f"{property_column}_Censors": f"{columnames[index]}_Censors",
            },
        )
    msg = (
        "No columns were converted"
        if np.array_equal(property_columns, columnames)
        else f"{property_columns} Columns where converted to the "
        f"following columns: {columnames}"
    )
    logger.info(msg)
    return columnames, dataframe


def _remove_too_big_molecules(
    per_compound_dataframe: pd.DataFrame,
    max_heavy: int,
) -> pd.DataFrame:
    """Remove all compounds that have num heavy atoms > max heavy.

    Args:
        per_compound_dataframe (pd.DataFrame): dataframe to remove
        compounds from
        max_heavy (int): max heavy atom threshold
    """
    before = len(per_compound_dataframe)
    per_compound_dataframe = per_compound_dataframe.loc[
        per_compound_dataframe["Num_Heavy_Atoms"] <= max_heavy
    ]
    after = len(per_compound_dataframe)
    logger.info(
        "Removing %i compounds with more than %i atoms from the dataset.",
        before - after,
        max_heavy,
    )
    return per_compound_dataframe


def parse_input_file(  # pylint: disable=R0913,W0102
    infile: Path,
    property_columns: list[str],
    error_path: Path,
    max_heavy: int,
    units: Sequence[str | None] = DEFAULT_UNIT,
    delimiter: str = "tab",
    series_column: str | None = None,
    verbose: bool = False,
) -> tuple[pd.DataFrame, list[str]]:
    """Parse and cleanup input file for nonadditivity analysis.

    Reads the infile and converts it to the dataframe. subsequently it renames the
    columns,adds molecular weight and nr of heavy atoms to the entries,
    removes duplicate entries, and removes entries where the smiles was not parseable.
    It also checks and converts any activity values to the
    negative log10 of their molar value.

    Args:
        infile (str): path to the input file.
        error_path (str): path to file, where the faulty smiles are written to.
        property_columns (list[str]): Name(s) of the property columns to be converted.
        units (Union[str, list[str]]): Unit(s) of the values in the property column(s).
        defaults to "M"
        max_heavy (int): max number of heavy atoms per ligand.
        verbose (bool, optional): True for verbose output. Defaults to None.
        delimiter (str | None, optional): delimiter in inputfile. Defaults to None.
        series_column (str | None, optional): name of series column. Defaults to None.

    Returns:
        tuple[pd.DataFrame, list[str]]: dataframe and names of the
        columns with the converted activity values.
    """
    # Read infile and convert to df
    input_dataframe = _read_in_data(
        infile=infile,
        property_columns=property_columns,
        delimiter=delimiter,
        series_column=series_column,
    )

    # add mol weight, nur heavy atoms and get list of faulty/unknown smiles
    (
        unknown_smiles_indices,
        faulty_smiles_indices,
        input_dataframe,
    ) = _parse_input_smiles(
        dataframe=input_dataframe,
        verbose=verbose,
    )

    # remove entries with said smiles from dataframe
    input_dataframe = _remove_unknown_and_faulty_structures(
        dataframe=input_dataframe,
        unknown_smiles_indices=unknown_smiles_indices,
        faulty_smiles_indices=faulty_smiles_indices,
        error_path=error_path,
    )

    input_dataframe = _check_for_censored_values(
        dataframe=input_dataframe,
        property_columns=property_columns,
    )

    # remove entries with duplicate structures
    input_dataframe = _remove_duplicate_structures(dataframe=input_dataframe)

    # check and convert activity values to molar -log10.
    conv_activity_rows, input_dataframe = _check_and_convert_activities(
        dataframe=input_dataframe,
        property_columns=property_columns,
        units=units,
    )

    input_dataframe = _remove_too_big_molecules(
        per_compound_dataframe=input_dataframe,
        max_heavy=max_heavy,
    )

    return input_dataframe, conv_activity_rows
